# **********************************************************************
# Copyright 2014 VMware, Inc.  All rights reserved. VMware Confidential.
# **********************************************************************
# $Id$
# $DateTime$
# $Change$
# $Author$
# **********************************************************************
import nsx_logging
import os
import json
import uuid

import cli.commands.host_shared.ovs_util as ovs_util

from netaddr import IPNetwork
from cli.commands.cli_command_provider import CliCommandProvider


lg = nsx_logging.getLogger("cli.commands.host_shared.logical_router")

class UuidArgument(object):
    @staticmethod
    def verify(uuid_str, req_dict):
        try:
            uuid_val = uuid.UUID(uuid_str)
        except ValueError:
            return False
        return True

class HostGetLogicalRouter(CliCommandProvider):
    def __call__(self, cmd_list):
        res = {}
        ports = []
        lr_uuid = cmd_list[2]
        lr_id = ovs_util.lr_uuid_to_id(lr_uuid)
        if lr_id is None:
            return 'MSG_LR_NOT_FOUND'
        port_strs = ovs_util.get_lif_list(lr_uuid)
        for port_str in port_strs:
            ports.append({'prefix': port_str[1],
                          'mac': port_str[2]})
        curr_neigh = ovs_util.lr_uuid_to_curr_neighbors(lr_uuid)
        max_neigh = ovs_util.lr_uuid_to_max_neighbors(lr_uuid)
        res['uuid'] = lr_uuid
        res['id'] = lr_id
        res['ports'] = ports
        res['curr_neigh'] = curr_neigh
        res['max_neigh'] = max_neigh
        return "OK\n\n%s" % json.dumps(res)

class HostGetLogicalRouterPorts(CliCommandProvider):
    def __call__(self, cmd_list):
        res = []
        lr_uuid = cmd_list[2]
        lr_id = ovs_util.lr_uuid_to_id(lr_uuid)
        if lr_id is None:
            return 'MSG_LR_NOT_FOUND'
        args = [lr_uuid]
        if len(cmd_list) >= 5:
            args.append(cmd_list[4])
        ports = ovs_util.get_lif_list(*args)
        lif_dhcp_relay_map = ovs_util.get_lr_dhcp_relay_info(lr_uuid)
        lif_ra_config_map = ovs_util.get_lr_ra_config(lr_uuid)
        if lif_ra_config_map is None:
            return 'MSG_LR_PORT_RA_CONFIG_NOT_FOUND'
        lif_dad_config_map = ovs_util.get_lr_dad_config(lr_uuid)
        if lif_dad_config_map is None:
            return 'MSG_LR_PORT_DAD_CONFIG_NOT_FOUND'
        for port in ports:
            lif = lif_dhcp_relay_map[port[0]]
            dhcp_relay_info = []
            if lif[1] == 'Enabled':
                dhcp_relay_info.extend(lif[2:])
            else:
                dhcp_relay_info.append('Not enabled')
            lif_ra_config = lif_ra_config_map[port[0]]
            # if RA is enabled, it will contain M=0/1 and O=0/1 flags.
            if len(lif_ra_config) > 3 and "O=" in lif_ra_config[2]:
                ra_mode = lif_ra_config[1:3]
            else:
                ra_mode = lif_ra_config[1:2]
            lif_dad_config = lif_dad_config_map[port[0]]
            dad_mode = lif_dad_config[1]
            res.append({'uuid': port[0], 'prefix':port[1],
                        'mac': port[2], 'vni': int(port[3]), 'flag': port[4],
                        'dhcp_relay' : dhcp_relay_info,
                        'ra_mode' : ra_mode,
                        'dad_mode' : dad_mode})
        return "OK\n\n%s" % json.dumps(res)

class HostGetLogicalRouterPortStats(CliCommandProvider):
    def __call__(self, cmd_list):
        version = None
        v4_res = []
        v6_res = []
        lr_uuid = cmd_list[2]
        lr_id = ovs_util.lr_uuid_to_id(lr_uuid)
        if lr_id is None:
            return 'MSG_LR_NOT_FOUND'
        lif = cmd_list[4]
        if len(cmd_list) > 6:
            version = cmd_list[-1]
        cmd_stats = [
                ('lif/ingress', 'IPv4 Packets RX', 'IPv6 Packets RX'),
                ('lif/egress', 'IPv4 Packets TX', 'IPv6 Packets TX'),
                ('lif/ingress-drops', 'IPv4 RX Packets Drops',
                 'IPv6 RX Packets Drops'),
                ('lif/egress-drops', 'IPv4 TX Packets Drops',
                 'IPv6 TX Packets Drops'),
                ('lif/ttl-drops', 'IPv4 TTL Zero Drops',
                 'IPv6 TTL Zero Drops'),
                ('lif/urpf-drops', 'IPv4 uRPF Drops', 'IPV6 uRPF Drops'),
        ]
        # Common stats for IPV4 and IPV6
        for cmd, v4_st_name, v6_st_name in cmd_stats:
            st = ovs_util.run_exporter_cli(cmd, lif, split=True, header=True)
            if (version is None or version == "ipv4"):
                v4_res.append({'name': v4_st_name, 'counter': st[0][1]})
            if (version is None or version == "ipv6"):
                v6_res.append({'name': v6_st_name, 'counter': st[0][2]})
        # RS and RA packet stats is applicable only for IPV6
        if (version is None or version == "ipv6"):
            st = ovs_util.run_exporter_cli("lif/rs-ra-redirect-stats", lif,
                                           split=True, header=True)
            v6_res.append({'name': "RS RX Packets", 'counter': st[0][1]})
            v6_res.append({'name': "RA TX Packets", 'counter': st[0][2]})
            v6_res.append({'name': "Redirect TX Packets", 'counter': st[0][3]})
        return "OK\n\n%s" % json.dumps(v4_res + v6_res)

class HostGetLogicalRouterRoutes(CliCommandProvider):
    def __call__(self, cmd_list):
        res = []
        lr_uuid = cmd_list[2]
        lr_id = ovs_util.lr_uuid_to_id(lr_uuid)
        if lr_id is None:
            return 'MSG_LR_NOT_FOUND'
        args = [lr_uuid]
        if len(cmd_list) > 4:
            args.append(cmd_list[4])
        routes = ovs_util.get_lr_route(args)
        for route in routes:
            res.append({'prefix': route[0], 'gateway': route[1],
                        'lif': route[2], 'status': route[3]})
        return "OK\n\n%s" % json.dumps(res)

class HostGetLogicalRouterNeighbors(CliCommandProvider):
    def __call__(self, cmd_list):
        res = []
        lr_uuid = cmd_list[2]
        lr_id = ovs_util.lr_uuid_to_id(lr_uuid)
        if lr_id is None:
            return 'MSG_LR_NOT_FOUND'
        ip = None
        if len(cmd_list) > 4:
            ip = cmd_list[4]
        neighbors = ovs_util.get_lr_neighbors(lr_uuid)
        for neighbor in neighbors:
            if ip is None or IPNetwork(neighbor[2]) == IPNetwork(ip):
                res.append({'prefix': str(IPNetwork(neighbor[2])),
                            'mac': neighbor[3], 'state': neighbor[4],
                            'timeout': neighbor[5]})
        return "OK\n\n%s" % json.dumps(res)
