# **********************************************************************
# Copyright 2021 VMware, Inc.  All rights reserved. VMware Confidential.
# **********************************************************************
import collections
import json
import logging
import nsx_logging
import os
import re
import sys
import time
import hashlib

# Always suppress the same message within the RESET_INTERVAL
# There is a possibility of the duplicate log missing
REPEATS = "REPEATS:"
CONFIG_JSON_FILE = "/opt/vmware/nsx-common/python/nsx_logging/config.json"
LRU_DICT_SIZE = 100
RESET_INTERVAL_IN_SEC = 10


def generate_id(cur_info):
    return hashlib.sha256(cur_info.encode('utf-8')).hexdigest()


class LRUDict(object):
    def __init__(self, size):
        self.ordered_dict = collections.OrderedDict()
        self.size = size

    def get(self, key):
        return self.ordered_dict.get(key, None)

    def set(self, key, count, cur_time, record, record_tag_dict):
        deleted_record = None
        if len(self.ordered_dict) >= self.size:
            deleted_record = self.ordered_dict.popitem(last=False)[1]
        first_log_time = cur_time
        # try to get first_log_time from dict if exists
        if key in self.ordered_dict:
            first_log_time = self.ordered_dict.get(key)[1]
        self.ordered_dict[key] = (count, first_log_time
                , record, record_tag_dict)
        return deleted_record

    def getList(self):
        return list(self.ordered_dict.values())


class DuplicateFilter(logging.Filter):
    def __init__(self, config_json = CONFIG_JSON_FILE):
        try:
            with open(config_json, 'r') as load_f:
                load_dict = json.load(load_f)
                lru_dict_size = load_dict\
                    .get('duplicate_filter_lru_dict_size', LRU_DICT_SIZE)
                reset_interval_in_sec = load_dict\
                    .get('duplicate_filter_reset_interval_in_sec', RESET_INTERVAL_IN_SEC)
                self.lru_dict_size = lru_dict_size
                self.reset_interval_in_sec = reset_interval_in_sec
        except:
            self.lru_dict_size = LRU_DICT_SIZE
            self.reset_interval_in_sec = RESET_INTERVAL_IN_SEC
        self.lru_dict = LRUDict(self.lru_dict_size)
        self.last_interval = time.time()
        #add lock for multi-process
        if sys.platform == "win32":
            from threading import Lock
            self.lock = Lock()
        else:
            try:
                from gevent.coros import BoundedSemaphore
                self.lock = BoundedSemaphore(1)
            except ImportError:
                from threading import Lock
                self.lock = Lock()


    def print_suppressed_log_message(self, cur_record_tuple, cur_time):
        cur_count = cur_record_tuple[0]
        if cur_count == 0:
            return
        first_log_time = cur_record_tuple[1]
        cur_record = cur_record_tuple[2]
        cur_level = cur_record.levelno
        cur_log_name = cur_record.name
        cur_log_msg = cur_record.msg
        cur_record_args = cur_record.args
        cur_record_tags_dict = cur_record_tuple[3]
        lg = nsx_logging.getLogger(cur_log_name)
        lg.log(cur_level, "{header} {num} repeats in {time} sec: {msg}"\
            .format(header=REPEATS,
            num=cur_count,
            time=int(cur_time - first_log_time),
            msg=cur_log_msg) % cur_record_args, **cur_record_tags_dict)


    def filter(self, record):
        if type(record.msg) is str and record.msg.startswith(REPEATS):
            return True
        cur_time = time.time()

        record_tag_dict = {}
        if type(record.__dict__.get("tags")) == str:
            cur_tags = re.split(r"[ ]+", record.__dict__.get("tags"))
            for cur_tag in cur_tags:
                cur_tag_list = cur_tag.split("=")
                if len(cur_tag_list) == 2:
                    record_tag_dict[cur_tag_list[0]] = cur_tag_list[1][1:-1]
        elif type(record.__dict__.get("tags")) == dict:
            record_tag_dict = record.__dict__.get("tags")
        #special case, in case arg == [], change it to ""
        #example:
        #obj.setdefault("lockout_immune_addresses",[])
        #lg.info("lockout_immune_addresses from DB  %s",obj["lockout_immune_addresses"])
        record_id = generate_id(str(record.levelno) + str(record.name)
            + str(record.msg) + str(record.args) + str(sorted(record_tag_dict)))
        # If this is the first log line of each reset_interval, iterate the current lru_dict map
        # and print out all the counter > 0 logs in the map. After that, reset the lru_dict map
        # and print out current log line.
        if cur_time - self.last_interval >= self.reset_interval_in_sec:
            # make the call synchronized
            self.lock.acquire()
            cur_list = []
            try:
                cur_list = self.lru_dict.getList()
                self.lru_dict = LRUDict(size=self.lru_dict_size)
                self.lru_dict.set(record_id,
                    0, cur_time, record, record_tag_dict)
                self.last_interval = cur_time
            finally:
                self.lock.release()
            for cur_tuple in cur_list:
                self.print_suppressed_log_message(cur_tuple, cur_time)
            return True
        is_return_current_record = False
        deleted_record_tuple = None
        # make the call synchronized
        self.lock.acquire()
        try:
            lru_dict_value = self.lru_dict.get(record_id)
            # If the current log line is already in the lru_dict map, increase its counter
            # and not print out this current log line.
            if lru_dict_value != None:
                last_count = lru_dict_value[0]
                deleted_record_tuple = self.lru_dict.set(record_id,
                    last_count + 1, cur_time, record, record_tag_dict)
                is_return_current_record = False
            # If the current log line is not in the lru_dict map, add this entry in the map
            # and print out this current log line.
            else:
                deleted_record_tuple = self.lru_dict.set(record_id,
                    0, cur_time, record, record_tag_dict)
                is_return_current_record = True
        finally:
            self.lock.release()
        if deleted_record_tuple is not None:
            self.print_suppressed_log_message(deleted_record_tuple, cur_time)
        return is_return_current_record

# -----------------------------------------------------------------------------
_DUPLICATE_FILTER = DuplicateFilter()
