# Copyright (C) 2021 VMware, Inc. All rights reserved.
# -- VMware Confidential
import subprocess
from sha.contrib.metric.abc_base import Metric
from sha.contrib.metric.utils import run_command, timeout_cache
from sha import config
from sha.config import NODE_TYPE, DIST_ESX, DIST_KVM, DIST_BMS, SYSTEM, \
    PLATFORM_LINUX, LINUX_ID, LINUX_ID_CENTOS, LINUX_ID_RHEL, LINUX_VERSION_ID, \
    RHEL_VERSION_ID_76, RHEL_VERSION_ID_80, BUILD_VERSION, VERSION_ESX_67, VERSION_ESX_65
import re


from sha.preconfig import get_sub_logger

_logger = get_sub_logger(__name__, s2comp="metric-collector-NestdbStatus")


class NestdbStatus(Metric):

    @staticmethod
    def is_min_space_20kb():
        """
        Check whether the available disk space has a minimum value 20kb on current system
        (on some linux, the disk always try to keep 20KB space at least).
        """
        if SYSTEM != PLATFORM_LINUX:
            return False

        if LINUX_ID == LINUX_ID_CENTOS:
            # currently, tested on centos 7.6, 7.7, 7.8, 8.0, 8.2,
            # all above minimum disk space is 20KB
            #
            # QE and Jarvis does not have below setup. We can only infer their behaviors basing on above tested ones.
            #  KVM	CentOS 8.4  (use same logic with CentOS 8.2)
            #  KVM	CentOS 7.9  (use same logic with CentOS 7.8)
            return True

        if LINUX_ID == LINUX_ID_RHEL and LINUX_VERSION_ID not in {RHEL_VERSION_ID_76, RHEL_VERSION_ID_80}:
            # currently, only tested on redhat 7.6, 7.7, 7.8, 7.9, 8.0, 8.2.
            # for 7.6 and 8.0 minimum is 0KB. Others, are 20KB
            #
            # QE and Jarvis does not have below setup. We can only infer its behavior basing on above tested ones.
            # KVM	RHEL Server release 8.4 (use same logic with CentOS 8.2)
            return True

        # tested on OEL 7.7, 7.6 the minimum space is 0KB
        # supported hypervisors : http://rtqa-automation.eng.vmware.com:5000/supported_hypervisors

        return False

    @timeout_cache(2)
    def usage(self, *args, **kwargs):
        # ================ check nestdb process status ================
        nestdb_id_cmd = ['pidof', '-xs', '/opt/vmware/nsx-nestdb/bin/nestdb-server']
        try:
            output = run_command(nestdb_id_cmd)
        except subprocess.CalledProcessError as e:
            return_code = e.returncode
            if return_code == 1:
                # if nestdb is exit, should get error response as below. This is as expected.
                # Command '['pidof', '-xs', '/opt/vmware/nsx-nestdb/bin/nestdb-server']' returned non-zero exit status 1
                _logger.debug("Failed in get pid of nestdb, return code is 1")
                output = None
            else:
                raise e

        if output is not None:
            nestdb_is_running = True
            _logger.debug("nestdb is running. pid is {}".format(output))
        else:
            nestdb_is_running = False
            _logger.debug("nestdb is not running.")

        nestdb_watchdog_cmd_linux = ['ps', '-ef']
        nestdb_watchdog_cmd_esx = ['ps', '-c']

        if NODE_TYPE == DIST_ESX:
            nestdb_watchdog_cmd = nestdb_watchdog_cmd_esx
        elif NODE_TYPE in {DIST_KVM, DIST_BMS}:
            nestdb_watchdog_cmd = nestdb_watchdog_cmd_linux
        else:
            raise Exception("Invalid node type :{}".format(NODE_TYPE))

        # ================ check watchdog status ================
        output = run_command(nestdb_watchdog_cmd)
        match_object = re.search(r'watchdog\.sh -s NSX-NESTDB', output)
        if match_object is not None:
            watchdog_is_running = True
            _logger.debug("watchdog is running.")
        else:
            watchdog_is_running = False
            _logger.debug("watchdog is not running.")

        # ================ check disk usage ================
        disk_is_full = False
        nestdb_path = '/var/lib/vmware/nsx/nestdb'

        disk_usage_cmd = ['df', '-k', nestdb_path]
        output = run_command(disk_usage_cmd)
        # valid output example for cmd: df -k /var/lib/vmware/nsx/nestdb
        #
        # Filesystem          1K-blocks    Used Available Use% Mounted on
        # /dev/mapper/cl-root  29128812 3297968  25830844  12% /

        if len(output.splitlines()) == 2:
            _logger.debug("Got valid output:{} for cmd:{}".format(output, disk_usage_cmd))
            maximum = output.splitlines()[1].split()[1]
            used = output.splitlines()[1].split()[2]
            available = output.splitlines()[1].split()[3]
            use = output.splitlines()[1].split()[4][:-1]
            if int(available) == 0 and int(use) == 100:
                _logger.debug("disk path: {} is full. maximum:{} KiB used:{} KiB, available:{} KiB, use:{}%"
                              .format(nestdb_path, maximum, used, available, use))
                disk_is_full = True
            elif self.is_min_space_20kb() and int(available) <= 20 and int(use) == 100:
                # Treat less than 20KB space is full, that is because on some linux (e.g. centos and redhat77) the disk
                # always try to keep 20KB space at least.
                disk_is_full = True
                _logger.debug("disk path: {} is full. maximum:{} KiB used:{} KiB, "
                              "available:{} KiB, use:{}%".format(nestdb_path, maximum, used, available, use))
            else:
                _logger.debug("disk path: {} is not full. maximum:{} KiB used:{} KiB, available:{} KiB, use:{}%"
                              .format(nestdb_path, maximum, used, available, use))
        elif NODE_TYPE == DIST_ESX:
            # /var/lib/vmware/nsx/nestdb is not mounted on this esx, use localcli to check disk usage.
                disk_usage_cmd = ['localcli', 'system', 'visorfs', 'ramdisk', 'list']  # then check Ramdisk for var path
                output = run_command(disk_usage_cmd)
                _logger.debug("Got output:{} for cmd:{}".format(output, disk_usage_cmd))
                # example output for 'localcli system visorfs ramdisk list'
                #
                # Ramdisk Name  System  Include in Coredumps  Reserved   Maximum      Used       Peak Used  Free   Reserved Free  Maximum Inodes  Allocated Inodes  Used Inodes  Mount Point
                # ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------
                # root          true    true                  32768 KiB    32768 KiB   8384 KiB   8384 KiB   74 %           74 %           18080              6656         6632  /
                # etc           true    true                  28672 KiB    28672 KiB    272 KiB    280 KiB   99 %           99 %            4096              1024          769  /etc
                # opt           true    true                      0 KiB    32768 KiB   1632 KiB   1632 KiB   95 %            0 %            8192              1024          735  /opt
                # var           true    true                   5120 KiB    49152 KiB  11108 KiB  17568 KiB   77 %            0 %            8192              1248          810  /var
                # tmp           false   false                  2048 KiB   262144 KiB     40 KiB  33568 KiB   99 %           98 %            8192               256           10  /tmp
                # iofilters     false   false                     0 KiB    32768 KiB      0 KiB      0 KiB  100 %            0 %           10240                32            1  /var/run/iofilters
                # shm           false   false                     0 KiB  1048576 KiB      0 KiB      0 KiB  100 %            0 %             512                32            1  /var/run/shm
                # hostdstats    false   false                     0 KiB   388096 KiB   2728 KiB   2728 KiB   99 %            0 %            8192                32            4  /var/lib/vmware/hostd/stats
                # nsx-idps      false   false                     0 KiB    65536 KiB  10420 KiB  10420 KiB   84 %            0 %            8192                32            2  /etc/nsx-idps/rules
                match_object = re.search(r'var.*/var', output)
                maximum = match_object.group().split()[5]
                used = match_object.group().split()[7]
                free = match_object.group().split()[11]
                if int(maximum) == int(used) and int(free) == 0:
                    _logger.debug("disk path: {} is full. maximum:{} KiB, used:{} KiB, free:{}%"
                                  .format(nestdb_path, maximum, used, free))
                    disk_is_full = True
                else:
                    _logger.debug("disk path: {} is not full. maximum:{} KiB used:{} KiB free:{}%"
                                  .format(nestdb_path, maximum, used, free))
        else:
            _logger.warn("Not a valid output:{} for disk_usage_cmd:{} on node_type:{}"
                         .format(output, disk_usage_cmd, NODE_TYPE))

        return {"nestdb_is_running": nestdb_is_running,
                'watchdog_is_running': watchdog_is_running,
                'disk_is_full': disk_is_full}
