# Copyright (C) 2019-2021 VMware, Inc. All rights reserved.
# -- VMware Confidential

import gc
import traceback

import sha.config
from sha.core.channel.provider.base import BaseProvider
from sha.constants import DATA, WAVEFRONT_TYPE, WAVEFRONT_TYPE_METRIC, WAVEFRONT_TYPE_TRACING, \
                          COLLECTOR_HOST, COLLECTOR_METRICS_PORT, COLLECTOR_TRACING_PORT

from threading import Lock, Thread

from sha.core.channel.provider.wavefront_proxy_client import WavefrontProxyClient
from sha.core.channel.provider.firewall_ruleset import SHAFirewall

class WfProvider(BaseProvider):
    """The wavefront Provider"""
    def __init__(self):
        self._name = "WavefrontProvider"
        self._logger = sha.config.get_logger(__name__)

    def sync(self, msg):
        try:
            self._logger.debug("before sending metric to WavefrontSender, msg:{}".format(msg))
            if WavefrontSender.isReady():
                if (msg[WAVEFRONT_TYPE] == WAVEFRONT_TYPE_METRIC):
                    WavefrontSender.send_metric_now(msg[DATA])
                elif (msg[WAVEFRONT_TYPE] == WAVEFRONT_TYPE_TRACING):
                    WavefrontSender.send_span_now(msg[DATA])
                else:
                    raise Exception('Unknown wavefront type: {}'.format(msg[WAVEFRONT_TYPE]))
                self._logger.debug('Wavefront data sent {}'.format(msg))
            else:
                self._logger.debug('wavefront collector is not ready!')
        except Exception as e:
            trace = traceback.format_exc()
            self._logger.warning("{} failed in sending metric:{}, Exception:{}, trace:{}"
                                 .format(self.name, msg, e, trace))

    def close(self):
        WavefrontSender.clear()

    @property
    def name(self):
        return self._name


class WavefrontSender:
    """The wavefront sender.
    """
    _logger = sha.config.get_logger(__name__)
    _wavefront_sender = None
    _wavefront_conf = None
    _wavefront_lock = Lock()
    _wavefront_firewall = SHAFirewall("wavefront", "0301", "nsx-sha");


    @classmethod
    def _close_client(cls, stale_wavefront_sender):
        try:
            cls._logger.info('Close stale wavefront client: {}'.format(stale_wavefront_sender))
            stale_wavefront_sender.close()
            cls._logger.info('Stale wavefront client closed: {}'.format(stale_wavefront_sender))
            del stale_wavefront_sender
        except Exception as e:
            cls._logger.warning('Failed to close stale wavefront client:{}'.format(e))

    @classmethod
    def _clear(cls):
        with cls._wavefront_lock:
            cls._logger.info('Stale wavefront client: {}'.format(cls._wavefront_sender))
            stale_client = cls._wavefront_sender
            if cls._wavefront_sender is not None:
                cls._wavefront_sender = None

        if stale_client is not None:
            try:
                # Fire another thread to close stale wavefront client
                Thread(target=cls._close_client, args=[stale_client]).start()
            except Exception as e:
                cls._logger.warning('Failed to fire a thread to close stale wavefront client:{}'.format(e))

    @classmethod
    def update(cls, wavefront_conf):
        # update wavefront conf
        cls._logger.info('Update wavefront conf: {}'.format(wavefront_conf))
        cls._wavefront_conf = wavefront_conf

        # clear stale wavefront client
        cls._clear()

        # Create new wavefront client
        cls._logger.info('Create wavefront client to export data:')
        with cls._wavefront_lock:
            try:
                cls._wavefront_sender = WavefrontProxyClient(host=cls._wavefront_conf.get(COLLECTOR_HOST, None),
                                                             metrics_port=cls._wavefront_conf.get(COLLECTOR_METRICS_PORT, None),
                                                             tracing_port=cls._wavefront_conf.get(COLLECTOR_TRACING_PORT, None))
                cls._logger.info('Wavefront client created: {}'.format(cls._wavefront_sender))
            except Exception as e:
                cls._logger.warning('Failed to create wavefront client:{}'.format(e))
                cls._wavefront_sender = None

        # Update firewall rule
        cls._wavefront_firewall.update([{"protocol": "tcp", "port": cls._wavefront_conf.get(key)} \
                                        for key in [COLLECTOR_METRICS_PORT, COLLECTOR_TRACING_PORT] if cls._wavefront_conf.get(key, None)
                                       ])


    @classmethod
    def clear(cls):
        # Clear wavefront conf
        cls._wavefront_conf = None

        # clear stale wavefront client
        cls._clear()

        # update firewall rule
        cls._wavefront_firewall.update([])

    @classmethod
    def isReady(cls):
        ret = False
        if cls._wavefront_lock.acquire(timeout = 1):
            if cls._wavefront_sender is not None:
                ret = True
            cls._wavefront_lock.release()
        else:
            cls._logger.warning('Failed to acquire lock to check wavefront client')
        return ret

    @classmethod
    def send_metric_now(cls, wavefront_data):
        if not cls._wavefront_lock.acquire(timeout = 1):
            cls._logger.warning('Failed to acquire lock before sending wavefront data')
            return
        try:
            if cls._wavefront_sender is None:
                cls._logger.warning('Wavefront collector has gone')
            else:
                cls._wavefront_sender.send_metric_now(wavefront_data)
                cls._logger.debug('Batch wavefront metric data sent: {}'.format(wavefront_data))
        except Exception as e:
            trace = traceback.format_exc()
            cls._logger.warning('Failed to send batch wavefront metric sapn: {}\n {}'.format(e, trace))
            cls._logger.debug("Collecting...")
            try:
                n = gc.collect()
            except Exception as e:
                cls._logger.warning('Failed to collect garbage: {}'.format(e))
            cls._logger.debug("Number of unreachable objects collected by GC: {}".format(n))
            cls._logger.debug("Uncollectable garbage:{}".format(gc.garbage))
        finally:
            cls._wavefront_lock.release()

    @classmethod
    def send_span_now(cls, wavefront_data):
        if not cls._wavefront_lock.acquire(timeout = 1):
            cls._logger.warning('Failed to acquire lock before sending wavefront data')
            return
        try:
            if cls._wavefront_sender is None:
                cls._logger.warning('Wavefront collector has gone')
            else:
                cls._wavefront_sender.send_span_now(wavefront_data)
                cls._logger.debug('Batch wavefront tracing span sent: {}'.format(wavefront_data))
        except Exception as e:
            trace = traceback.format_exc()
            cls._logger.warning('Failed to send batch wavefront tracing sapn: {}\n {}'.format(e, trace))
            cls._logger.debug("Collecting...")
            try:
                n = gc.collect()
            except Exception as e:
                cls._logger.warning('Failed to collect garbage: {}'.format(e))
            cls._logger.debug("Number of unreachable objects collected by GC: {}".format(n))
            cls._logger.debug("Uncollectable garbage:{}".format(gc.garbage))
        finally:
            cls._wavefront_lock.release()
