# Copyright (C) 2019-2021 VMware, Inc. All rights reserved.
# -- VMware Confidential

import queue
import threading
import traceback
import sha.config
import sha.preconfig
from sha.constants import ENABLE, NORMAL, ERROR, DATA
import sha.constants
from sha.core.plugin.plugin_status_collection import PluginStatusCollection
from sha.core.profile._resolver import Resolver
from sha.core.whiteboard import DataWhiteboard


class MonitorNotifier(threading.Thread):
    """
    MonitorNotifier is responsible to excute plugin when the event
    generated from monitor
    to plugins which are listening for the monitor.
    """

    def __init__(self, plugin, *args, **kwargs):
        super().__init__(name='nsx-sha-monitor-notifier', *args, **kwargs)
        self._keep_running = True
        self._logger = sha.preconfig.get_sub_logger(__name__, s2comp='monitor-notifier-[{}]'.format(plugin.name))
        self._plugin = plugin
        self._notification_queue = queue.Queue(100)
        profile = self._plugin.load_profile()
        self.pre_unreached_msg = self._plugin.check_pre_condition()
        self.pre_enable = profile[ENABLE]
        self._logger.info("Set initial status to NORMAL for Monitor plugins.")
        PluginStatusCollection.update_plugin_status(plugin.id, NORMAL, '')

    def _handle_event(self, notification):
        try:
            profile = self._plugin.load_profile()
            enable = profile[ENABLE]
            unreached_msg = self._plugin.check_pre_condition()
            self.pre_unreached_msg = unreached_msg
            self.pre_enable = enable

            if not enable:
                PluginStatusCollection.update_plugin_status(self._plugin.id, NORMAL, 'Plugin is disabled.')
                self._logger.debug('plugin:{} is disabled, not handle notification:{}'
                                   .format(self._plugin.name, notification))
                return

            if unreached_msg is not None:
                PluginStatusCollection.update_plugin_status(self._plugin.id, NORMAL,
                                                            'Plugin is not executed as: {}'.format(unreached_msg))
                self._logger.debug('precondition of plugin:{} is not reached: {}, not handle notification:{}'
                                   .format(self._plugin.name, unreached_msg, notification))
                return

            self._plugin.execute(notification)
            self._logger.debug('plugin {} is executed with notification:{}'.format(self._plugin.name, notification))
            PluginStatusCollection.update_plugin_status(self._plugin.id, NORMAL, '')
        except Exception as e:
            PluginStatusCollection.update_plugin_status(self._plugin.id, ERROR, "Failed to run due to {}".format(str(e)))
            trace = traceback.format_exc()
            self._logger.error("Prepare plugin {} failed, exception:{}, trace:{}".format(self._plugin.name, e, trace))

    def run(self):
        self._logger.info("starting monitor notifier for: {}.".format(self._plugin.name))

        # Register events to data whiteboard
        DataWhiteboard.register(self._plugin.id, self._plugin.metric_names, self._notification_queue,
                                prefix_metrics=self._plugin.prefix_metric_names)
        while self._keep_running:
            notification = None
            try:
                try:
                    notification = DataWhiteboard.get_data(self._notification_queue, timeout=1)
                except queue.Empty:
                    if not self.need_check():
                        continue

                    notification = self.check_profile_and_condition()
                    if notification is None:
                        continue

                self._logger.debug("received notification: {}".format(notification))
                self._handle_event(notification)
            except Exception:
                t = traceback.format_exc()
                self._logger.error("Failed to dispatch notification:{}, trace:{}".format(notification, t))

        self._logger.info("stopped monitor notifier for: {}.".format(self._plugin.name))

    def check_profile_and_condition(self):
        profile = self._plugin.load_profile()
        cur_enable = profile[ENABLE]
        cur_unreached_msg = self._plugin.check_pre_condition()

        pre_running = self.pre_enable is True and self.pre_unreached_msg is None
        cur_running = cur_enable is True and cur_unreached_msg is None

        if not pre_running and cur_running:
            notification = DataWhiteboard.get_metrics(self._plugin.id, self._plugin.metric_names,
                                                      prefix_metrics=self._plugin.prefix_metric_names)
            self._logger.info("check_profile_and_condition - perform notify. pre_enable:{}, cur_enable:{}, "
                              "pre_unreached_msg:{}, cur_unreached_msg:{}, sent metrics:{} to plugin id:{}."
                              .format(self.pre_enable, cur_enable, self.pre_unreached_msg, cur_unreached_msg,
                                      notification,
                                      self._plugin.id))
            return notification
        else:
            self._logger.debug("check_profile_and_condition - no need to notify. pre_enable:{}, cur_enable:{}, "
                               "pre_unreached_msg:{}, cur_unreached_msg:{}"
                               .format(self.pre_enable, cur_enable, self.pre_unreached_msg, cur_unreached_msg))
            return None

    def need_check(self):
        """
        For those notifications only sent when metrics changed, plugin should check current metrics after profile is
        changed to enable or pre-condition is changed to reach. Otherwise, those notifications when plugin
        disabled or pre-condition unreached will be lost.
        """
        return self._plugin.id in {Resolver.UPLINK_STATUS_MONITOR_ID, Resolver.PROCESS_STATUS_MONITOR_ID}

    def stop(self):
        self._logger.warning("going to stop monitor notifier")
        self._keep_running = False
